#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module containing helper methods for creating DICOM datasets."""

import datetime
from typing import Callable, List, Optional, Sequence, Tuple, Union

from highdicom.content import (
    SpecimenDescription,
    SpecimenPreparationStep,
    SpecimenSampling,
    SpecimenStaining,
)
from pydicom.dataset import Dataset
from pydicom.sequence import Sequence as DicomSequence
from pydicom.sr.coding import Code
from pydicom.uid import UID as Uid
from pydicom.uid import generate_uid
from wsidicom.conceptcode import (
    AnatomicPathologySpecimenTypesCode,
    ConceptCode,
    SpecimenEmbeddingMediaCode,
    SpecimenFixativesCode,
    SpecimenSamplingProcedureCode,
    SpecimenStainsCode,
)

from opentile import Metadata


def create_base_dataset(
    modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
) -> Dataset:
    """Create a base dataset by combining module datasets with a minimal
    wsi dataset.

    Parameters
    ----------
    modules: Optional[Union[Dataset, Sequence[Dataset]]] = None

    Returns
    ----------
    Dataset
        Combined base dataset.
    """
    base_dataset = create_wsi_dataset()
    if modules is None:
        modules = create_default_modules()
    if isinstance(modules, Sequence):
        for module in modules:
            base_dataset.update(module)
    elif isinstance(modules, Dataset):
        base_dataset.update(modules)
    else:
        raise TypeError("datasets parameter should be single or list of Datasets")
    return base_dataset


def populate_base_dataset(
    metadata: Metadata, base_dataset: Dataset, include_confidential: bool = True
) -> Dataset:
    """Populate dataset with properties from tiler, if present.
    Parameters
    ----------
    metadata: Metadata
        A object holding metadata.
    base_dataset: Dataset
        Dataset to append properties to.
    include_confidential: bool = True
        If to include confidential properties (see https://dicom.nema.org/medical/dicom/current/output/html/part15.html#table_E.1-1)  # NOQA

    Returns
    ----------
    Dataset
        Dataset with added properties.
    """

    properties = {
        "Manufacturer": metadata.scanner_manufacturer,
        "ManufacturerModelName": metadata.scanner_model,
        "SoftwareVersions": metadata.scanner_software_versions,
    }
    confidential_properties = {
        "AcquisitionDateTime": metadata.aquisition_datetime,
        "DeviceSerialNumber": metadata.scanner_serial_number,
    }
    if include_confidential:
        properties.update(confidential_properties)
    for property_name, property_value in metadata.properties.items():
        if property_name == "lossy_image_compression_method":
            properties["LossyImageCompressionMethod"] = property_value
        elif property_name == "lossy_image_compression_ratio":
            properties["LossyImageCompressionRatio"] = property_value

    for property_name, property_value in properties.items():
        if not hasattr(base_dataset, property_name) and property_value is not None:
            setattr(base_dataset, property_name, property_value)

    return base_dataset


def create_wsi_dataset(uid_generator: Callable[..., Uid] = generate_uid) -> Dataset:
    """Return dataset containing (parts of) SOP common, general series, whole
    slide microscopy series, frame of reference, acquisition context,
    multi-frame dimension, and whole slide microscopy image modules.

    Some modules returned 'not complete', and completed during image data
    import or file save():
        SOP common module:
            SOPInstanceUID
        Whole slide microscopy image module:
            ImageType
            Aquisition DateTime
            SamplesPerPixel
            PhotometricInterpretation
            PlanarConfiguration
            NumberOfFrames
            BitsAllocated
            BitsStored
            HighBit
            PixelRepresentation
            LossyImageCompression (and conditionals)
            ImagedVolumeWidth
            ImagedVolumeWidth
            ImagedVolumeDepth
            TotalPixelMatrixColumns
            TotalPixelMatrixRows
            FocusMethod
            ExtendedDepthOfField (and conditionals)

    Parameters
    ----------
    uid_generator: Callable[..., Uid] = generate_uid
        Function that can generate Uids.

    Returns
    ----------
    Dataset
        WSI dataset.
    """
    dataset = Dataset()

    # SOP common module (SOPInstanceUID written on save())
    dataset.SOPClassUID = "1.2.840.10008.5.1.4.1.1.77.1.6"

    # dataset.StudyInstanceUID = uid_generator()
    # General series and Whole slide Microscopy modules
    dataset.SeriesNumber = ""
    dataset.SeriesInstanceUID = uid_generator()
    dataset.Modality = "SM"
    dataset.LossyImageCompression = "00"

    # Frame of reference module
    dataset.FrameOfReferenceUID = uid_generator()
    dataset.PositionReferenceIndicator = "SLIDE_CORNER"

    # Acquisition context module (empty)
    dataset.AcquisitionContextSequence = DicomSequence()

    # Multi-frame Dimension module
    dimension_organization_uid = uid_generator()
    dimension_organization_sequence = Dataset()
    dimension_organization_sequence.DimensionOrganizationUID = (
        dimension_organization_uid
    )
    dataset.DimensionOrganizationSequence = DicomSequence(
        [dimension_organization_sequence]
    )

    # Whole slide micropscopy image module (most filled when importing file)
    dataset.BurnedInAnnotation = "NO"
    dataset.SpecimenLabelInImage = "NO"
    dataset.VolumetricProperties = "VOLUME"
    # AcquisitionDateTime is required
    dt = datetime.datetime.now()
    dataset.AcquisitionDateTime = dt.strftime("%Y%m%d%H%M%S.%f")

    return dataset


def create_study_module(
    study_id: str = "",
    study_date: Optional[datetime.date] = None,
    study_time: Optional[datetime.time] = None,
    study_accession_number: str = "",
    referring_physician_name: str = "",
    uid_generator: Callable[..., Uid] = generate_uid,
) -> Dataset:
    """Create general study module.

    Parameters
    ----------
    study_id: str = ""
        Study identifier. Can be empty.
    Optional[datetime.date] = None
        Date the study started. Can be empty.
    Optional[datetime.time] = None
        Time the study started. Can be empty.
    accession_number: str = ""
        Order for the study. Can be empty.
    referring_physician_name: str = ""
        Name of reffering physician. Can be empty.

    Returns
    ----------
    Dataset
        Dataset containing general study module.
    """
    dataset = Dataset()
    dataset.StudyInstanceUID = uid_generator()
    dataset.StudyID = study_id
    dataset.StudyDate = study_date
    dataset.StudyTime = study_time
    dataset.AccessionNumber = study_accession_number
    dataset.ReferringPhysicianName = referring_physician_name
    return dataset


def create_patient_module(
    name: str = "",
    id: str = "",
    birth_date: Optional[datetime.date] = None,
    sex: str = "",
    identity_removed: bool = False,
    de_indentification_methods: Optional[Sequence[Union[str, Code]]] = None,
    age_at_extraction: Optional[str] = None,
) -> Dataset:
    """Create patient and patient study modules.

    Parameters
    ----------
    name: str = ""
        Patient's full name. Can be empty.
    id: str = ""
        Primary identifier for the Patient. Can be empty.
    birth_date: str = ""
        Birth date of the Patient. Can be empty.
    sex: str = ""
        Sex of the named Patient. F/M/O. Can be empty.
    identity_removed: bool = False
        True if partient idendity has been removed from attributes in dataset
        and in image data.
    de_indentification_methods: Optional[Sequence[Union[str, Code]]] = None
        Method of de-indentification. Can be descriptive strings and/or coded
        values.
    age_at_extraction: Optional[str] = None
        Age of the Patient at extraction.

    Returns
    ----------
    Dataset
        Dataset containing patient and patient study modules.
    """
    if de_indentification_methods is None:
        de_indentification_methods = []
    dataset = Dataset()
    dataset.PatientName = name
    dataset.PatientID = id
    dataset.PatientBirthDate = birth_date
    dataset.PatientSex = sex
    dataset.PatientIdentityRemoved = "YES" if identity_removed else "NO"
    if identity_removed and de_indentification_methods is None:
        raise ValueError(
            "de-intification method must be specified if "
            "patient identity has been removed"
        )
    de_indentification_method_strings: List[str] = []
    de_indentification_method_codes: List[Code] = []

    for de_indentification_method in de_indentification_methods:
        if isinstance(de_indentification_method, str):
            de_indentification_method_strings.append(de_indentification_method)
        elif isinstance(de_indentification_method, Code):
            de_indentification_method_codes.append(de_indentification_method)
        else:
            raise TypeError(
                "De-indentification methods should be string or coded value"
            )
    if de_indentification_method_strings != []:
        dataset.DeidentificationMethod = de_indentification_method_strings
    if de_indentification_method_codes != []:
        dataset.DeidentificationMethodCodeSequence = DicomSequence(
            [
                ConceptCode.from_code(de_indentification_method_code).to_ds()
                for de_indentification_method_code in de_indentification_method_codes
            ]
        )
    if age_at_extraction is not None:
        dataset.PatientAge = age_at_extraction

    return dataset


def create_device_module(
    manufacturer: str = "Unknown",
    model_name: str = "Unknown",
    serial_number: str = "Unknown",
    software_versions: Sequence[str] = ["Unknown"],
) -> Dataset:
    """Create extended equipment module.

    Parameters
    ----------
    manufacturer: str = 'Unknown'.
        Manufacturer of the equipment.
    model_name: str = 'Unknown'
        Manufacturer's model name of the equipment.
    serial_number: str = 'Unknown'
        Manufacturer's serial number of the equipment.
    software_versions: Sequence[str] = ['Unknown']
        Software version of the equipment.

    Returns
    ----------
    Dataset
        Dataset containing extended equipment module.
    """
    dataset = Dataset()
    properties = {
        "Manufacturer": manufacturer,
        "ManufacturerModelName": model_name,
        "DeviceSerialNumber": serial_number,
        "SoftwareVersions": software_versions,
    }
    for name, value in properties.items():
        setattr(dataset, name, value)
    return dataset


def create_sample(
    sample_id: str,
    stainings: Optional[Sequence[str]] = None,
    embedding_medium: Optional[str] = None,
    fixative: Optional[str] = None,
    specimen_id: Optional[str] = None,
    specimen_type: Optional[str] = None,
    specimen_sampling_method: Optional[str] = None,
    anatomical_sites: Optional[Sequence[Tuple[Code, Sequence[Code]]]] = None,
    location: Optional[Union[str, Tuple[float, float, float]]] = None,
    uid_generator: Callable[..., Uid] = generate_uid,
) -> Dataset:
    """Create sample description.

    Parameters
    ----------
    sample_id: str
        Identifier for the sample.
    stainings: Optional[str] = None
        Stainings used. See SpecimenStainsCode.list() for allowed values.
    embedding_medium: Optional[str] = None
        Embedding medium used. See SpecimenEmbeddingMediaCode.list() for
        allowed values.
    fixative: Optional[str] = None
        Fixative used. See SpecimenFixativesCode.list() for allowed values.
    specimen_id: Optional[str] = None
        Identifier for the specimen the sample was sampled from.
    specimen_type: Optional[str] = None
        Anatotomic type of specimen the sample was sampled from. See
        AnatomicPathologySpecimenTypesCode.list() for allowed values.
    specimen_sampling_method: Optional[str] = None
        Sampling method used for sampling the sample from the specimen. See
        SpecimenSamplingProcedureCode.list() for allowed values.
    anatomical_sites: Optional[Sequence[Tuple[Code, Sequence[Code]]]] = None
        List of original anatomical sites, each defined by a code and optional
        list of modifier codes.
    location: Optional[Union[str, Tuple[float, float, float]]] = None
        Location of sample in slide, defined either by descriptive string
        or xyz-coordinates.
    uid_generator: Callable[..., Uid] = generate_uid
        Function that can generate uids.

    Returns
    ----------
    Dataset
        Dataset containing a sample description.
    """
    sample_preparation_steps: List[SpecimenPreparationStep] = []
    if stainings is not None:
        sample_preparation_step = create_sample_preparation_step(
            sample_id, stainings, embedding_medium, fixative
        )
        sample_preparation_steps.append(sample_preparation_step)
    if specimen_id is not None:
        sample_sampling_step = create_sample_sampling_step(
            sample_id, specimen_id, specimen_sampling_method, specimen_type
        )
        sample_preparation_steps.append(sample_sampling_step)

    specimen = SpecimenDescription(
        specimen_id=sample_id,
        specimen_uid=uid_generator(),
        specimen_preparation_steps=sample_preparation_steps,
        specimen_location=location,
    )
    if anatomical_sites is not None:
        specimen = add_anatomical_sites_to_specimen(specimen, anatomical_sites)

    return specimen


def create_sample_preparation_step(
    specimen_id: str,
    stainings: Sequence[str],
    embedding_medium: Optional[str],
    fixative: Optional[str],
) -> SpecimenPreparationStep:
    """Create SpecimenPreparationStep for a preparation step.

    Parameters
    ----------
    specimen_id: str
        ID of specimen that has been prepared.
    stainings: Sequence[str]
        Sequence of stainings used.
    embedding_medium: Optional[str] = None
        Embedding medium used.
    fixative: Optional[str] = None
        Fixative used.

    Returns
    ----------
    SpecimenPreparationStep
        SpecimenPreparationStep for a preparation step.
    """
    if embedding_medium is not None:
        embedding_medium_code = SpecimenEmbeddingMediaCode(embedding_medium).code
    else:
        embedding_medium_code = None
    if fixative is not None:
        fixative_code = SpecimenFixativesCode(fixative).code
    else:
        fixative_code = None

    processing_procedure = SpecimenStaining(
        [SpecimenStainsCode(staining).code for staining in stainings]
    )
    sample_preparation_step = SpecimenPreparationStep(
        specimen_id=specimen_id,
        processing_procedure=processing_procedure,
        embedding_medium=embedding_medium_code,
        fixative=fixative_code,
    )
    return sample_preparation_step


def create_sample_sampling_step(
    sample_id: str,
    specimen_id: str,
    specimen_sampling_method: Optional[str],
    specimen_type: Optional[str],
) -> SpecimenPreparationStep:
    """Create SpecimenPreparationStep for a sampling step.

    Parameters
    ----------
    sample_id: str
        ID of sample that has been created.
    specimen_id: str
        ID of specimen that was sampled.
    specimen_sampling_method: Optional[str]
        Method used to sample specimen.
    specimen_type: Optional[str]
        Type of specimen that was sampled

    Returns
    ----------
    SpecimenPreparationStep
        SpecimenPreparationStep for a sampling step.
    """
    if specimen_sampling_method is None:
        raise ValueError(
            "Specimen sampling method required if " "specimen id is defined"
        )
    if specimen_type is None:
        raise ValueError("Specimen type required if " "specimen id is defined")
    specimen_type_code = AnatomicPathologySpecimenTypesCode(specimen_type)
    sampling_method_code = SpecimenSamplingProcedureCode(specimen_sampling_method)
    sample_sampling_step = SpecimenPreparationStep(
        specimen_id=sample_id,
        processing_procedure=SpecimenSampling(
            method=sampling_method_code.code,
            parent_specimen_id=specimen_id,
            parent_specimen_type=specimen_type_code.code,
        ),
    )
    return sample_sampling_step


def add_anatomical_sites_to_specimen(
    specimen: Dataset, anatomical_sites: Sequence[Tuple[Code, Sequence[Code]]]
) -> Dataset:
    """Add anatomical site and anatomical site modifier codes to specimen.

    Parameters
    ----------
    specimen: Dataset
        Dataset containing a specimen description
    anatomical_sites: Sequence[Tuple[Code, Sequence[Code]]]
        List of original anatomical sites, each defined by a code and
        list of modifier codes (can be empty).

    Returns
    ----------
    Dataset
        Dataset containing a specimen description.
    """
    anatomical_site_datasets: List[Dataset] = []
    for (anatomical_site, modifiers) in anatomical_sites:
        anatomical_site_dataset = ConceptCode.from_code(anatomical_site).to_ds()

        if modifiers != []:
            modifier_datasets: List[Dataset] = []
            for modifier in modifiers:
                modifier_dataset = Dataset()
                modifier_dataset.CodeValue = modifier.value
                modifier_dataset.CodingSchemeDesignator = modifier.scheme_designator
                modifier_dataset.CodeMeaning = modifier.meaning
                modifier_datasets.append(modifier_dataset)

            (
                anatomical_site_dataset.PrimaryAnatomicStructureModifierSequence
            ) = DicomSequence(modifier_datasets)
        anatomical_site_datasets.append(anatomical_site_dataset)
    specimen.PrimaryAnatomicStructureSequence = DicomSequence(anatomical_site_datasets)
    return specimen


def create_specimen_module(
    slide_id: str, samples: Union[Dataset, Sequence[Dataset]]
) -> Dataset:
    """Create specimen module.

    Parameters
    ----------
    slide_id: str.
        Identifier for the slide.
    samples: Union[Dataset, Sequence[Dataset]]
        Single or list of sample descriptions (should be created with
        create_sample())

    Returns
    ----------
    Dataset
        Dataset containing specimen module.
    """
    dataset = Dataset()
    dataset.ContainerIdentifier = slide_id
    dataset.IssuerOfTheContainerIdentifierSequence = DicomSequence()

    container_type_code_sequence = Dataset()
    container_type_code_sequence.CodeValue = "258661006"
    container_type_code_sequence.CodingSchemeDesignator = "SCT"
    container_type_code_sequence.CodeMeaning = "Slide"
    dataset.ContainerTypeCodeSequence = DicomSequence([container_type_code_sequence])

    container_component_sequence = Dataset()
    container_component_sequence.ContainerComponentMaterial = "GLASS"
    container_component_type_code_sequence = Dataset()
    container_component_type_code_sequence.CodeValue = "433472003"
    container_component_type_code_sequence.CodingSchemeDesignator = "SCT"
    container_component_type_code_sequence.CodeMeaning = "Microscope slide coverslip"
    container_component_sequence.ContainerComponentTypeCodeSequence = DicomSequence(
        [container_component_type_code_sequence]
    )
    dataset.ContainerComponentSequence = DicomSequence([container_component_sequence])
    if not isinstance(samples, Sequence):
        samples = [samples]
    specimen_description_sequence = DicomSequence(samples)
    dataset.SpecimenDescriptionSequence = specimen_description_sequence

    return dataset


def create_brightfield_optical_path_module(
    magnification: Optional[float] = None,
) -> Dataset:
    """Create optical path module for brightfield illumination conditions.

    Parameters
    ----------
    magnification: Optional[float] = None
        Objective magnification.

    Returns
    ----------
    Dataset
        Dataset containing optical path module for brightfield illumination.
    """
    dataset = Dataset()
    optical_path_item = Dataset()
    optical_path_item.OpticalPathIdentifier = "0"
    illumination_type_code_sequence = Dataset()
    illumination_type_code_sequence.CodeValue = "111744"
    illumination_type_code_sequence.CodingSchemeDesignator = "DCM"
    illumination_type_code_sequence.CodeMeaning = "Brightfield illumination"
    optical_path_item.IlluminationTypeCodeSequence = DicomSequence(
        [illumination_type_code_sequence]
    )
    illumination_color_code_sequence = Dataset()
    illumination_color_code_sequence.CodeValue = "R-102C0"
    illumination_color_code_sequence.CodingSchemeDesignator = "SRT"
    illumination_color_code_sequence.CodeMeaning = "Full Spectrum"
    optical_path_item.IlluminationColorCodeSequence = DicomSequence(
        [illumination_color_code_sequence]
    )
    if magnification is not None:
        optical_path_item.ObjectiveLensPower = magnification

    dataset.OpticalPathSequence = DicomSequence([optical_path_item])

    return dataset


def create_default_modules(
    uid_generator: Callable[..., Uid] = generate_uid
) -> List[Dataset]:
    """Return default module dataset for testing.

    Parameters
    ----------
    uid_generator: Callable[..., Uid]
        Function that can generate Uids.

    Returns
    ----------
    List[Dataset]
        Default module datasets.
    """
    modules: List[Dataset] = []

    # Generic study module
    modules.append(create_study_module())

    # Generic patient module
    modules.append(create_patient_module())

    # Generic device module
    modules.append(create_device_module())

    # Generic specimen module
    modules.append(
        create_specimen_module(
            "Unknown",
            samples=[
                create_sample(
                    sample_id="Unknown",
                    stainings=["water soluble eosin stain", "hematoxylin stain"],
                    uid_generator=uid_generator,
                )
            ],
        )
    )

    # Generic optical path sequence
    modules.append(create_brightfield_optical_path_module())

    return modules



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source for reading opentile compatible file."""

from pathlib import Path
from typing import List, Optional, Sequence, Union

from opentile import OpenTile
from opentile.metadata import Metadata
from pydicom import Dataset

from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.encoding import Encoder
from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.sources.opentile.opentile_image_data import OpenTileImageData


class OpenTileSource(DicomizerSource):
    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: int = 512,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
    ) -> None:
        self._tiler = OpenTile.open(filepath, tile_size)
        self._metadata = self._tiler.metadata
        super().__init__(
            filepath,
            encoder,
            tile_size,
            modules,
            include_levels,
            include_label,
            include_overview,
            include_confidential,
        )

    def close(self):
        self._tiler.close()

    @property
    def has_label(self) -> bool:
        return len(self._tiler.labels) > 0

    @property
    def has_overview(self) -> bool:
        return len(self._tiler.overviews) > 0

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @property
    def pyramid_levels(self) -> List[int]:
        return [level.pyramid_index for level in self._tiler.levels]

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        """Return True if file in filepath is supported by OpenTile."""
        return OpenTile.detect_format(filepath) is not None

    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        level = self._tiler.levels[level_index]
        return OpenTileImageData(level, self._encoder, self.metadata.image_offset)

    def _create_label_image_data(self) -> DicomizerImageData:
        label = self._tiler.labels[0]
        return OpenTileImageData(label, self._encoder)

    def _create_overview_image_data(self) -> DicomizerImageData:
        overview = self._tiler.overviews[0]
        return OpenTileImageData(overview, self._encoder)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Image data for tiffslide compatible file."""

import math
import re
from enum import Enum
from typing import List, Optional, Tuple

import numpy as np
from PIL import Image
from PIL.Image import Image as PILImage
from pydicom.uid import UID as Uid
from tiffslide import TiffSlide
from tiffslide.tiffslide import (
    PROPERTY_NAME_BACKGROUND_COLOR,
    PROPERTY_NAME_BOUNDS_HEIGHT,
    PROPERTY_NAME_BOUNDS_WIDTH,
    PROPERTY_NAME_BOUNDS_X,
    PROPERTY_NAME_BOUNDS_Y,
    PROPERTY_NAME_MPP_X,
    PROPERTY_NAME_MPP_Y,
)
from wsidicom.errors import WsiDicomNotFoundError
from wsidicom.geometry import Orientation, Point, PointMm, Region, Size, SizeMm
from wsidicom.instance import ImageCoordinateSystem

from wsidicomizer.encoding import Encoder
from wsidicomizer.image_data import DicomizerImageData


class TiffSlideAssociatedImageType(Enum):
    LABEL = "label"
    MACRO = "macro"


class TiffSlideImageData(DicomizerImageData):
    def __init__(self, tiff_slide: TiffSlide, encoder: Encoder):
        """Wraps a TiffSlide image to ImageData.

        Parameters
        ----------
        tiff_slide: TiffSlide
            TiffSlide object to wrap.
        encoded: Encoder
            Encoder to use.
        """
        super().__init__(encoder)
        self._slide = tiff_slide
        self._blank_color = self._get_blank_color(self.photometric_interpretation)

    @property
    def transfer_syntax(self) -> Uid:
        """The uid of the transfer syntax of the image."""
        return self._encoder.transfer_syntax

    @property
    def photometric_interpretation(self) -> str:
        return self._encoder.photometric_interpretation(self.samples_per_pixel)

    @property
    def samples_per_pixel(self) -> int:
        return 3

    @property
    def focal_planes(self) -> List[float]:
        return [0.0]

    @property
    def optical_paths(self) -> List[str]:
        return ["0"]

    @property
    def blank_color(self) -> Tuple[int, int, int]:
        return self._blank_color

    def _get_blank_color(self, photometric_interpretation: str) -> Tuple[int, int, int]:
        """Return color to use blank tiles. Parses background color from
        tiffslide if present.

        Parameters
        ----------
        photometric_interpretation: str
            The photomoetric interpretation of the dataset

        Returns
        ----------
        Tuple[int, int, int]
            RGB color.

        """
        slide_background_color_string = self._slide.properties.get(
            PROPERTY_NAME_BACKGROUND_COLOR
        )
        if slide_background_color_string is not None:
            rgb = re.findall(r"([0-9a-fA-F]{2})", slide_background_color_string)
            if len(rgb) == 3:
                return (int(rgb[0], 16), int(rgb[1], 16), int(rgb[2], 16))
        return super()._get_blank_color(photometric_interpretation)


class TiffSlideAssociatedImageData(TiffSlideImageData):
    def __init__(
        self,
        tiff_slide: TiffSlide,
        image_type: TiffSlideAssociatedImageType,
        encoder: Encoder,
    ):
        """Wraps a TiffSlide associated image (label or overview) to ImageData.

        Parameters
        ----------
        tiff_slide: TiffSlide
            TiffSlide object to wrap.
        image_type: TiffSlideAssociatedImageType
            Type of image to wrap.
        encoded: Encoder
            Encoder to use.
        """
        super().__init__(tiff_slide, encoder)
        self._image_type = image_type
        if image_type.value not in self._slide.associated_images:
            raise ValueError(f"{image_type.value} not in {self._slide}")

        image = self._slide.associated_images[image_type.value]
        self._image_size = Size.from_tuple(image.size)
        self._decoded_image = image
        self._encoded_image = self._encode(np.asarray(image))

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self.image_size

    @property
    def pixel_spacing(self) -> Optional[SizeMm]:
        """Size of the pixels in mm/pixel."""
        return None

    @property
    def pyramid_index(self) -> int:
        """The pyramidal index in relation to the base layer."""
        return 0

    def _get_encoded_tile(self, tile: Point, z: float, path: str) -> bytes:
        if tile != Point(0, 0):
            raise ValueError("Point(0, 0) only valid tile for non-tiled image")
        return self._encoded_image

    def _get_decoded_tile(self, tile: Point, z: float, path: str) -> PILImage:
        if tile != Point(0, 0):
            raise ValueError("Point(0, 0) only valid tile for non-tiled image")
        return self._decoded_image


class TiffSlideLevelImageData(TiffSlideImageData):
    def __init__(
        self, tiff_slide: TiffSlide, level_index: int, tile_size: int, encoder: Encoder
    ):
        super().__init__(tiff_slide, encoder)
        """Wraps a TiffSlide level to ImageData.

        Parameters
        ----------
        tiff_slide: TiffSlide
            TiffSlide object to wrap.
        level_index: int
            Level in TiffSlide object to wrap
        tile_size: int
            Output tile size.
        encoded: Encoder
            Encoder to use.
        """
        self._tile_size = Size(tile_size, tile_size)
        self._slide = tiff_slide
        self._level_index = level_index
        self._image_size = Size.from_tuple(
            self._slide.level_dimensions[self._level_index]
        )
        self._downsample = self._slide.level_downsamples[self._level_index]
        self._pyramid_index = int(round(math.log2(self.downsample)))
        try:
            base_mpp_x = float(self._slide.properties[PROPERTY_NAME_MPP_X])
            base_mpp_y = float(self._slide.properties[PROPERTY_NAME_MPP_Y])
            self._pixel_spacing = SizeMm(
                base_mpp_x * self.downsample / 1000.0,
                base_mpp_y * self.downsample / 1000.0,
            )
        except KeyError:
            raise Exception(
                "Could not determine pixel spacing as tiffslide did not "
                "provide mpp from the file."
            )

        # Get set image origin and size to bounds if available
        bounds_x = self._slide.properties.get(PROPERTY_NAME_BOUNDS_X)
        bounds_y = self._slide.properties.get(PROPERTY_NAME_BOUNDS_Y)
        bounds_w = self._slide.properties.get(PROPERTY_NAME_BOUNDS_WIDTH)
        bounds_h = self._slide.properties.get(PROPERTY_NAME_BOUNDS_HEIGHT)
        if bounds_x is not None and bounds_y is not None:
            self._offset = Point(int(bounds_x), int(bounds_y))
        else:
            self._offset = Point(0, 0)
        if bounds_w is not None and bounds_h is not None:
            self._image_size = Size(int(bounds_w), int(bounds_h)) // int(
                round(self.downsample)
            )
        else:
            self._image_size = Size.from_tuple(
                self._slide.level_dimensions[self._level_index]
            )

        self._blank_encoded_frame = bytes()
        self._blank_encoded_frame_size = None
        self._blank_decoded_frame = None
        self._blank_decoded_frame_size = None
        self._image_coordinate_system = ImageCoordinateSystem(
            PointMm(
                self._offset.x * base_mpp_x / 1000, self._offset.y * base_mpp_y / 1000
            ),
            Orientation((0, 1, 0, 1, 0, 0)),
        )

    @property
    def image_size(self) -> Size:
        """The pixel size of the image."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """The pixel tile size of the image."""
        return self._tile_size

    @property
    def pixel_spacing(self) -> SizeMm:
        """Size of the pixels in mm/pixel."""
        return self._pixel_spacing

    @property
    def downsample(self) -> float:
        """Downsample facator for level."""
        return self._downsample

    @property
    def pyramid_index(self) -> int:
        """The pyramidal index in relation to the base layer."""
        return self._pyramid_index

    @property
    def image_coordinate_system(self) -> ImageCoordinateSystem:
        return self._image_coordinate_system

    def stitch_tiles(
        self, region: Region, path: str, z: float, threads: int
    ) -> PILImage:
        """Overrides ImageData stitch_tiles() to read reagion directly from
        tiffslide object.

        Parameters
        ----------
        region: Region
             Pixel region to stitch to image
        path: str
            Optical path
        z: float
            Z coordinate

        Returns
        ----------
        PILImage
            Stitched image
        """
        if z not in self.focal_planes:
            raise WsiDicomNotFoundError(f"focal plane {z}", str(self))
        if path not in self.optical_paths:
            raise WsiDicomNotFoundError(f"optical path {path}", str(self))
        image_data = self._get_region(region)
        if image_data is None:
            image_data = self._get_blank_decoded_frame(region.size)
        return Image.fromarray(image_data)

    def _detect_blank_tile(self, data: np.ndarray) -> bool:
        """Detect if tile data is a blank tile, i.e. either has full
        transparency or is filled with background color. First checks if the
        corners are transparent or has background color before checking whole
        data.

        Parameters
        ----------
        data: np.ndarray
            Data to check if blank.

        Returns
        ----------
        bool
            True if tile is blank.
        """

        TOP = RIGHT = -1
        BOTTOM = LEFT = 0
        CORNERS_Y = [BOTTOM, BOTTOM, TOP, TOP]
        CORNERS_X = [LEFT, RIGHT, LEFT, RIGHT]
        background = np.array(self.blank_color)
        corners_rgb = np.ix_(CORNERS_X, CORNERS_Y)
        if np.all(data[corners_rgb] == background):
            if np.all(data == background):
                return True
        return False

    def _get_blank_encoded_frame(self, size: Size) -> bytes:
        """Return cached blank encoded frame for size, or create frame if
        cached frame not available or of wrong size.

        Parameters
        ----------
        size: Size
            Size of frame to get.

        Returns
        ----------
        bytes
            Encoded blank frame.
        """
        if self._blank_encoded_frame_size != size:
            frame = np.full(
                size.to_tuple() + (3,), self.blank_color, dtype=np.dtype(np.uint8)
            )
            self._blank_encoded_frame = self._encode(frame)
            self._blank_encoded_frame_size = size
        return self._blank_encoded_frame

    def _get_blank_decoded_frame(self, size: Size) -> PILImage:
        """Return cached blank decoded frame for size, or create frame if
        cached frame not available or of wrong size.

        Parameters
        ----------
        size: Size
            Size of frame to get.

        Returns
        ----------
        bytes
            Decoded blank frame.
        """
        if self._blank_decoded_frame is None or self._blank_decoded_frame_size != size:
            frame = Image.new("RGB", size.to_tuple(), self.blank_color)
            self._blank_decoded_frame = frame
        return self._blank_decoded_frame

    def _get_region(self, region: Region) -> Optional[np.ndarray]:
        """Return Image read from region in tiffslide image. If image data for
        region is blank, None is returned. Transparent pixels are made into
        background color

        Parameters
        ----------
        region: Region
            Region to get image for.

        Returns
        ----------
        Optional[np.ndarray]
            Image data of region, or None if region is blank.
        """
        if region.size.width < 0 or region.size.height < 0:
            raise ValueError("Negative size not allowed")

        location_in_base_level = region.start * self.downsample + self._offset

        region_data = self._slide.read_region(
            location_in_base_level.to_tuple(),
            self._level_index,
            region.size.to_tuple(),
            as_array=True,
        )
        if self._detect_blank_tile(region_data):
            return None

        return region_data

    def _get_encoded_tile(self, tile_point: Point, z: float, path: str) -> bytes:
        """Return image bytes for tile. Transparency is removed and tile is
        encoded as jpeg.

        Parameters
        ----------
        tile_point: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        bytes
            Tile bytes.
        """
        if z not in self.focal_planes:
            raise WsiDicomNotFoundError(f"focal plane {z}", str(self))
        if path not in self.optical_paths:
            raise WsiDicomNotFoundError(f"optical path {path}", str(self))
        tile = self._get_region(Region(tile_point * self.tile_size, self.tile_size))
        if tile is None:
            return self._get_blank_encoded_frame(self.tile_size)
        return self._encode(tile)

    def _get_decoded_tile(self, tile_point: Point, z: float, path: str) -> PILImage:
        """Return Image for tile. Image mode is RGB.

        Parameters
        ----------
        tile_point: Point
            Tile position to get.
        z: float
            Focal plane of tile to get.
        path: str
            Optical path of tile to get.

        Returns
        ----------
        PILImage
            Tile as Image.
        """
        if z not in self.focal_planes:
            raise WsiDicomNotFoundError(f"focal plane {z}", str(self))
        if path not in self.optical_paths:
            raise WsiDicomNotFoundError(f"optical path {path}", str(self))
        tile = self._get_region(Region(tile_point * self.tile_size, self.tile_size))
        if tile is None:
            return self._get_blank_decoded_frame(self.tile_size)
        return Image.fromarray(tile)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module containing a base Source implementation suitable for use with non-DICOM
files."""

from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import List, Optional, Sequence, Union

from opentile.metadata import Metadata
from pydicom import Dataset, config
from pydicom.dataset import Dataset
from wsidicom.instance import ImageType, WsiDataset, WsiInstance
from wsidicom.source import Source
from wsidicom.graphical_annotations import AnnotationInstance

from wsidicomizer.dataset import create_base_dataset, populate_base_dataset
from wsidicomizer.encoding import Encoder
from wsidicomizer.image_data import DicomizerImageData

config.enforce_valid_values = True
config.future_behavior()


class DicomizerSource(Source, metaclass=ABCMeta):
    """
    Metaclass for Dicomizer sources. Subclasses should implement the method
    is_supported(), _create_level_image_data(), _create_label_image_data(), and
     _create_overview_image_data() and the properties metadata, pyramid_levels,
     has_label, and has_overview. Subclasses can override the __init__().
    """

    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: int = 512,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
    ) -> None:
        self._filepath = filepath
        self._encoder = encoder
        self._tile_size = tile_size
        self._modules = modules
        self._include_levels = include_levels
        self._include_label = include_label
        self._include_overview = include_overview
        self._include_confidential = include_confidential
        self._base_dataset = populate_base_dataset(
            self.metadata, create_base_dataset(modules), include_confidential
        )

    @staticmethod
    @abstractmethod
    def is_supported(path: Path) -> bool:
        """Return True if file in filepath is supported by Dicomizer."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def metadata(self) -> Metadata:
        """Return metadata for file."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def pyramid_levels(self) -> List[int]:
        """Return pyramid levels (scalings) for file."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def has_label(self) -> bool:
        """Return True if file has a label image."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def has_overview(self) -> bool:
        """Return True if file has a overview image."""
        raise NotImplementedError()

    @abstractmethod
    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        """Return image data instance for level."""
        raise NotImplementedError()

    @abstractmethod
    def _create_label_image_data(
        self,
    ) -> DicomizerImageData:
        """Return image data instance for label."""
        raise NotImplementedError()

    @abstractmethod
    def _create_overview_image_data(self) -> DicomizerImageData:
        """Return image data instance for overview."""
        raise NotImplementedError()

    @property
    def base_dataset(self) -> WsiDataset:
        return WsiDataset(self._base_dataset)

    @property
    def level_instances(self) -> List[WsiInstance]:
        return [
            WsiInstance.create_instance(
                self._create_level_image_data(level_index),
                self._base_dataset,
                ImageType.VOLUME,
            )
            for level_index in range(len(self.pyramid_levels))
            if self._is_included_level(
                self.pyramid_levels[level_index],
                self.pyramid_levels,
                self._include_levels,
            )
        ]

    @property
    def label_instances(self) -> List[WsiInstance]:
        if not self.has_label or not self._include_label:
            return []

        label = WsiInstance.create_instance(
            self._create_label_image_data(), self._base_dataset, ImageType.LABEL
        )
        return [label]

    @property
    def overview_instances(self) -> List[WsiInstance]:
        if not self.has_overview or not self._include_overview:
            return []

        overview = WsiInstance.create_instance(
            self._create_overview_image_data(), self._base_dataset, ImageType.OVERVIEW
        )
        return [overview]

    @property
    def annotation_instances(self) -> List[AnnotationInstance]:
        return []

    @staticmethod
    def _is_included_level(
        level: int,
        present_levels: Sequence[int],
        include_indices: Optional[Sequence[int]] = None,
    ) -> bool:
        """Return true if pyramid level is in included levels.

        Parameters
        ----------
        level: int
            Pyramid level to check.
        present_levels: Sequence[int]
            List of pyramid levels present.
        include_indices: Optional[Sequence[int]] = None
            Optional list indices (in present levels) to include, e.g. [0, 1]
            includes the two lowest levels. Negative indicies can be used,
            e.g. [-1, -2] includes the two highest levels. Default of None
            will not limit the selection. An empty sequence will exluded all
            levels.

        Returns
        ----------
        bool
            True if level should be included.
        """
        if level not in present_levels:
            return False
        if include_indices is None:
            return True
        absolute_levels = [
            present_levels[level]
            for level in include_indices
            if -len(present_levels) <= level < len(present_levels)
        ]
        return level in absolute_levels



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import argparse
import json
import os
from pathlib import Path
from typing import Optional, Sequence

from pydicom.dataset import Dataset

from wsidicomizer.wsidicomizer import WsiDicomizer


class WsiDicomizerCli:
    def __init__(self):
        self._parser = argparse.ArgumentParser(
            description=('Convert compatible wsi file to DICOM')
        )

        self._parser.add_argument(
            '-i', '--input',
            type=Path,
            required=True,
            help='Path to input wsi file.'
        )
        self._parser.add_argument(
            '-o', '--output',
            type=Path,
            help=(
                'Path to output folder. Folder will be created and must not '
                'excist. If not specified a folder named after the input file '
                'is created in the same path.'
            )
        )
        self._parser.add_argument(
            '-t', '--tile-size',
            type=int,
            default=512,
            help=(
                'Tile size (same for width and height). Required for ndpi and '
                'openslide formats E.g. 512'
            )
        )
        self._parser.add_argument(
            '-d', '--dataset',
            type=Path,
            help=(
                'Path to json DICOM dataset. Can be used to define additional '
                'DICOM modules to include in the files'
            )
        )
        self._parser.add_argument(
            '-l', '--levels',
            type=int,
            nargs='+',
            help=(
                'Pyramid levels to include, if not all. E.g. 0 1 for base and '
                'first pyramid layer. '
            ),
        )
        self._parser.add_argument(
            '--label',
            type=Path,
            help='Optional label image to use instead of label found in file.'
        )
        self._parser.add_argument(
            '--no-label',
            action="store_true",
            help='If not to include label'
        )
        self._parser.add_argument(
            '--no-overview',
            action="store_true",
            help='If not to include overview'
        )
        self._parser.add_argument(
            '--no-confidential',
            action="store_true",
            help='If not to include confidential metadata'
        )
        self._parser.add_argument(
            '-w', '--workers',
            type=int,
            default=os.cpu_count(),
            help='Number of worker threads to use'
        )
        self._parser.add_argument(
            '--chunk-size',
            type=int,
            default=100,
            help='Number of tiles to give each worker at a time'
        )
        self._parser.add_argument(
            '--format',
            type=str,
            default='jpeg',
            help="Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'."
        )
        self._parser.add_argument(
            '--quality',
            type=float,
            default=90,
            help=(
                "Quality to use if re-encoding. It is recommended to not use > 95 for "
                "jpeg. Use < 1 or > 1000 for lossless jpeg2000."
            )
        )
        self._parser.add_argument(
            '--subsampling',
            type=str,
            default='420',
            help=(
                "Subsampling option if using jpeg for re-encoding. Use '444' "
                "for no subsampling, '422' for 2x1 subsampling, and '420' for "
                "2x2 subsampling."
            )
        )
        self._parser.add_argument(
            '--offset-table',
            type=str,
            default='bot',
            help=(
                "Offset table to use, 'bot' basic offset table, 'eot' "
                "extended offset table, 'None' - no offset table."
            )
        )

    def cli(self):
        args = self._parser.parse_args()
        if not args.dataset:
            dataset = None
        else:
            json_file = open(args.dataset)
            dataset = Dataset.from_json(json.load(json_file))
        if not args.levels:
            levels = None
        else:
            levels = args.levels
        offset_table: Optional[str] = args.offset_table
        if offset_table == 'None':
            offset_table = None
        self.convert(
            filepath=args.input,
            output_path=args.output,
            modules=dataset,
            tile_size=args.tile_size,
            include_levels=levels,
            include_label=not args.no_label,
            include_overview=not args.no_overview,
            include_confidential=not args.no_confidential,
            workers=args.workers,
            chunk_size=args.chunk_size,
            encoding_format=args.format,
            encoding_quality=args.quality,
            jpeg_subsampling=args.subsampling,
            offset_table=offset_table,
            label=args.label
        )

    def convert(
        self,
        filepath: Path,
        output_path: Path,
        modules: Optional[Dataset] = None,
        tile_size: int = 512,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        workers: Optional[int] = None,
        chunk_size: Optional[int] = None,
        encoding_format: str = 'jpeg',
        encoding_quality: float = 90,
        jpeg_subsampling: str = '420',
        offset_table: Optional[str] = 'bot',
        label: Optional[Path] = None
    ):
        WsiDicomizer.convert(
            filepath=filepath,
            output_path=output_path,
            modules=modules,
            tile_size=tile_size,
            include_levels=include_levels,
            include_label=include_label,
            include_overview=include_overview,
            include_confidential=include_confidential,
            workers=workers,
            chunk_size=chunk_size,
            encoding_format=encoding_format,
            encoding_quality=encoding_quality,
            jpeg_subsampling=jpeg_subsampling,
            offset_table=offset_table,
            label=label
        )


def main():
    cli = WsiDicomizerCli()
    cli.cli()


if __name__ == "__main__":
    main()



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module containing a base ImageData implementation suitable for use with non-DICOM
files."""

from abc import ABCMeta, abstractmethod

import numpy as np
from wsidicom import ImageData
from wsidicom.geometry import Orientation, PointMm
from wsidicom.instance import ImageCoordinateSystem

from wsidicomizer.encoding import Encoder


class DicomizerImageData(ImageData, metaclass=ABCMeta):
    """
    Metaclass for Dicomizer image data. Subclasses should implement all the abstract
    methods and properties in the base ImageData-class, and the pyramid_index property.
    """

    _default_z = None

    def __init__(self, encoder: Encoder):
        """Metaclass for Dicomized image data.

        Parameters
        ----------
        encoded: Encoder
            Encoder to use.
        """
        self._encoder = encoder

    @property
    @abstractmethod
    def pyramid_index(self) -> int:
        """Should return pyramid level for image data."""
        raise NotImplementedError()

    @property
    def image_coordinate_system(self) -> ImageCoordinateSystem:
        """Return a default ImageCoordinateSystem."""
        return ImageCoordinateSystem(PointMm(0, 0), Orientation((0, 1, 0, 1, 0, 0)))

    def _encode(self, image_data: np.ndarray) -> bytes:
        """Return image data encoded in jpeg using set quality and subsample
        options.

        Parameters
        ----------
        image_data: np.ndarray
            Image data to encode.

        Returns
        ----------
        bytes
            Jpeg bytes.
        """
        return self._encoder.encode(image_data)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source using bioformats."""

from pathlib import Path
from typing import List, Optional, Sequence, Tuple, Union

from opentile.metadata import Metadata
from pydicom import Dataset

from wsidicomizer.dataset import create_base_dataset
from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.encoding import Encoder
from wsidicomizer.extras.bioformats.bioformats_image_data import \
    BioformatsImageData
from wsidicomizer.extras.bioformats.bioformats_reader import BioformatsReader
from wsidicomizer.image_data import DicomizerImageData


class BioformatsSource(DicomizerSource):
    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: Optional[int] = None,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        readers: Optional[int] = None,
        cache_path: Optional[str] = None
    ) -> None:
        if tile_size is None:
            raise ValueError("Tile size required for bioformats")
        self._base_dataset = create_base_dataset(modules)
        self._reader = BioformatsReader(
            Path(filepath),
            readers,
            cache_path
        )
        (
            self._pyramid_image_index,
            self._label_image_index,
            self._overview_image_index
        ) = self._get_image_indices(self._reader)
        super().__init__(
            filepath,
            encoder,
            tile_size,
            modules,
            include_levels,
            include_label,
            include_overview,
            include_confidential
        )

    @property
    def has_label(self) -> bool:
        return self._label_image_index is not None

    @property
    def has_overview(self) -> bool:
        return self._overview_image_index is not None

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        """Return True if file in filepath is supported by Bio-Formats."""
        return BioformatsImageData.detect_format(filepath)

    @property
    def pyramid_levels(self) -> List[int]:
        """Return pyramid levels (scalings) for file."""
        return list(range(self._reader.resolution_count(self._pyramid_image_index)))

    @property
    def metadata(self) -> Metadata:
        return Metadata()

    @staticmethod
    def _get_image_indices(
        reader: BioformatsReader
    ) -> Tuple[int, Optional[int], Optional[int]]:
        image_indices = list(range(reader.images_count))
        overview_image_index = None
        label_image_index = None

        for image_index in image_indices.copy():
            image_name = reader.image_name(image_index)
            if image_name is None:
                continue
            if (
                'macro' in image_name.lower()
                or 'overview' in image_name.lower()
            ):
                overview_image_index = image_index
                image_indices.remove(image_index)
            elif 'label' in image_name.lower():
                label_image_index = image_index
                image_indices.remove(image_index)

        pyramid_image_index = 0
        largest_image_width = None
        for image_index in image_indices:
            image_width = reader.size(image_index).width
            if (
                largest_image_width is None
                or largest_image_width < image_width
            ):
                pyramid_image_index = image_index
                largest_image_width = image_width
        return pyramid_image_index, label_image_index, overview_image_index

    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        return BioformatsImageData(
            self._reader,
            self._tile_size,
            self._encoder,
            self._pyramid_image_index,
            level_index
        )

    def _create_label_image_data(self) -> DicomizerImageData:
        assert self._label_image_index is not None
        return BioformatsImageData(
            self._reader,
            self._tile_size,
            self._encoder,
            self._label_image_index,
            0
        )

    def _create_overview_image_data(self) -> DicomizerImageData:
        assert self._overview_image_index is not None
        return BioformatsImageData(
            self._reader,
            self._tile_size,
            self._encoder,
            self._overview_image_index,
            0
        )

    def close(self) -> None:
        return self._reader.close()



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source for reading czi file."""

from pathlib import Path
from typing import List, Optional, Sequence, Union

from opentile.metadata import Metadata
from pydicom import Dataset

from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.encoding import Encoder
from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.sources.czi.czi_image_data import CziImageData


class CziSource(DicomizerSource):
    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: int = 512,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
    ) -> None:
        self._imaga_data = CziImageData(filepath, tile_size, encoder)
        self._metadata = self._imaga_data.metadata
        super().__init__(
            filepath,
            encoder,
            tile_size,
            modules,
            include_levels,
            include_label,
            include_overview,
            include_confidential,
        )

    def close(self) -> None:
        return self._imaga_data.close()

    @property
    def has_label(self) -> bool:
        return False

    @property
    def has_overview(self) -> bool:
        return False

    @property
    def pyramid_levels(self) -> List[int]:
        return [0]

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        """Return True if file in filepath is supported by CziFile."""
        return CziImageData.detect_format(filepath) is not None

    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        if level_index != 0:
            raise ValueError()  # TODO
        return CziImageData(self._filepath, self._tile_size, self._encoder)

    def _create_label_image_data(self) -> DicomizerImageData:
        return super()._create_label_image_data()

    def _create_overview_image_data(self) -> DicomizerImageData:
        return super()._create_overview_image_data()



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Cli for bioformats source."""

from pathlib import Path
from typing import Optional, Sequence

from pydicom import Dataset

from wsidicomizer.cli import WsiDicomizerCli
from wsidicomizer.extras.bioformats.bioformats_dicomizer import \
    BioformatsDicomizer


class BioformatsCli(WsiDicomizerCli):
    def convert(
        self,
        filepath: str,
        output_path: str,
        modules: Optional[Dataset] = None,
        tile_size: int = 512,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        workers: Optional[int] = None,
        chunk_size: Optional[int] = None,
        encoding_format: str = 'jpeg',
        encoding_quality: float = 90,
        jpeg_subsampling: str = '420',
        offset_table: Optional[str] = 'bot',
        label: Optional[Path] = None

    ):
        with BioformatsDicomizer.open(
            filepath=filepath,
            modules=modules,
            tile_size=tile_size,
            include_levels=include_levels,
            include_label=include_label,
            include_overview=include_overview,
            encoding_format=encoding_format,
            encoding_quality=encoding_quality,
            jpeg_subsampling=jpeg_subsampling,
            label=label
        ) as wsi:
            wsi.save(
                output_path=output_path,
                offset_table=offset_table,
                workers=workers,
                chunk_size=chunk_size
            )


def main():
    cli = BioformatsCli()
    cli.cli()


if __name__ == "__main__":
    main()



#    Copyright 2021 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Modulule containing image encodes that can be used to encode raw or decoded images
using implemented image encoders.
"""

from abc import ABCMeta, abstractmethod
from typing import Union, Optional
import numpy as np
from imagecodecs import jpeg2k_encode, jpeg8_encode
from pydicom.uid import UID, JPEGBaseline8Bit, JPEG2000Lossless, JPEG2000


class Encoder(metaclass=ABCMeta):
    """Abstract class for an image encoder."""

    @property
    @abstractmethod
    def transfer_syntax(self) -> UID:
        """Should return corresponding transfer syntax for encoded data."""
        raise NotImplementedError

    @property
    @abstractmethod
    def quality(self) -> Union[int, float]:
        """Should return quality setting of encoder."""
        raise NotImplementedError

    @abstractmethod
    def photometric_interpretation(self, channels: int) -> str:
        raise NotImplementedError()

    @abstractmethod
    def encode(self, data: np.ndarray) -> bytes:
        """Should return data as encoded bytes."""
        raise NotImplementedError()

    @classmethod
    def create_encoder(
        cls, format: str, quality: float, subsampling: Optional[str] = None
    ) -> "Encoder":
        """Creates an encoder with specified settings.

        Parameters
        ----------
        format: str
            Format for encoder, either 'jpeg' or 'jpeg2000.
        quality: float
            The encoding quality.
        subsampling: Optional[str] = None
            Subsampling setting (for jpeg).

        Returns
        ----------
        Enocer
            Encoder for settings.
        """
        if format == "jpeg":
            return JpegEncoder(quality=int(quality), subsampling=subsampling)
        elif format == "jpeg2000":
            return Jpeg2000Encoder(quality=quality)
        raise ValueError("Encoder format must be 'jpeg' or 'jpeg2000'")


class JpegEncoder(Encoder):
    """Encoder for JPEG."""

    def __init__(self, quality: int = 90, subsampling: Optional[str] = "420") -> None:
        """Creates a JPEG encoder with specified settings.

        Parameters
        ----------
        quality: int = 90
            The encoding quality. Recommended to not use higher than 95.
        subsampling: Optional[str] = '420'
            Subsampling option. Use '444' for no subsampling, '422' for 2x1
            subsampling, and '420' for 2x2 subsampling.
        """
        self._quality = quality
        if subsampling not in ["444", "422", "420"]:
            raise NotImplementedError(
                f"Only '444', '422' and '420' subsampling options implemented."
            )
        self._subsampling = subsampling

    @property
    def transfer_syntax(self) -> UID:
        """Transfer syntax of encoder."""
        return JPEGBaseline8Bit

    def photometric_interpretation(self, channels: int) -> str:
        if channels == 1:
            return "MONOCHROME2"
        elif channels == 3:
            return "YBR_FULL_422"
        raise ValueError()

    @property
    def quality(self) -> int:
        """Quality setting of encoder"""
        return self._quality

    @property
    def subsampling(self) -> Optional[str]:
        """Subsampling of encoder"""
        return self._subsampling

    def encode(self, data: np.ndarray) -> bytes:
        """Encodes data as JPEG. Converts data to uint8 before conversion.

        Parameters
        ----------
        data: np.ndarray
            Data to encode.

        Returns
        ----------
        bytes:
            JPEG bytes.
        """
        if data.dtype != np.dtype(np.uint8):
            data = (255 * (data / np.iinfo(data.dtype).max)).astype(np.uint8)
        return jpeg8_encode(data, level=self._quality, subsampling=self._subsampling)


class Jpeg2000Encoder(Encoder):
    def __init__(self, quality: float = 20.0) -> None:
        """Creates a JPEG2000 encoder with specified settings.

        Parameters
        ----------
        quality: float = 20.0.
            The encoding quality as peak signal to noise (PSNR). Use < 1 or > 1000 for
            lossless quality. Up to 60 gives acceptable results.

        """
        self._quality = quality
        if self.quality < 1 or self.quality > 1000:
            self._transfer_syntax = JPEG2000Lossless
        else:
            self._transfer_syntax = JPEG2000

    def photometric_interpretation(self, channels: int) -> str:
        if channels == 1:
            return "MONOCRHOME2"
        elif channels == 3:
            if self.transfer_syntax == JPEG2000Lossless:
                return "YBR_RCT"
            return "YBR_ICT"
        raise ValueError()

    @property
    def transfer_syntax(self) -> UID:
        """Transfer syntax of encoder."""
        return self._transfer_syntax

    @property
    def quality(self) -> float:
        """Quality setting of encoder"""
        return self._quality

    def encode(self, data: np.ndarray) -> bytes:
        """Encodes data as JPEG2000.

        Parameters
        ----------
        data: np.ndarray
            Numpy array of data to encode.

        Returns
        ----------
        bytes
            JPEG2000 bytes.
        """
        return jpeg2k_encode(
            data,
            level=self._quality,
            codecformat="J2K",
        )



#    Copyright 2023 SECTRA AB

#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Source for reading tiffslide compatible file."""

import math
from pathlib import Path
from typing import List, Optional, Sequence, Union

from opentile.metadata import Metadata
from pydicom import Dataset
from tiffslide import TiffSlide

from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.encoding import Encoder
from wsidicomizer.image_data import DicomizerImageData
from wsidicomizer.sources.tiffslide.tiffslide_image_data import (
    TiffSlideAssociatedImageData,
    TiffSlideAssociatedImageType,
    TiffSlideLevelImageData,
)
from wsidicomizer.sources.tiffslide.tiffslide_metadata import TiffSlideMetadata


class TiffSlideSource(DicomizerSource):
    def __init__(
        self,
        filepath: Path,
        encoder: Encoder,
        tile_size: int = 512,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        **source_args,
    ) -> None:
        self._tiffslide = TiffSlide(filepath, **source_args)
        self._metadata = TiffSlideMetadata(self._tiffslide)
        self._pyramid_levels = [
            int(round(math.log2(downsample)))
            for downsample in self._tiffslide.level_downsamples
        ]
        super().__init__(
            filepath,
            encoder,
            tile_size,
            modules,
            include_levels,
            include_label,
            include_overview,
            include_confidential,
        )

    def close(self):
        self._tiffslide.close()

    @property
    def has_label(self) -> bool:
        return "label" in self._tiffslide.associated_images

    @property
    def has_overview(self) -> bool:
        return "macro" in self._tiffslide.associated_images

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @property
    def pyramid_levels(self) -> List[int]:
        return self._pyramid_levels

    @staticmethod
    def is_supported(filepath: Path) -> bool:
        """Return True if file in filepath is supported by TiffSlide."""
        format = TiffSlide.detect_format(filepath)
        return format is not None

    def _create_level_image_data(self, level_index: int) -> DicomizerImageData:
        return TiffSlideLevelImageData(
            self._tiffslide, level_index, self._tile_size, self._encoder
        )

    def _create_label_image_data(self) -> DicomizerImageData:
        return TiffSlideAssociatedImageData(
            self._tiffslide, TiffSlideAssociatedImageType.LABEL, self._encoder
        )

    def _create_overview_image_data(self) -> DicomizerImageData:
        return TiffSlideAssociatedImageData(
            self._tiffslide, TiffSlideAssociatedImageType.MACRO, self._encoder
        )



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Metadata for tiffslide file."""

from typing import Optional

from opentile.metadata import Metadata
from tiffslide import TiffSlide
from tiffslide.tiffslide import PROPERTY_NAME_OBJECTIVE_POWER, PROPERTY_NAME_VENDOR


class TiffSlideMetadata(Metadata):
    def __init__(self, slide: TiffSlide):
        magnification = slide.properties.get(PROPERTY_NAME_OBJECTIVE_POWER)
        if magnification is not None:
            self._magnification = float(magnification)
        else:
            self._magnification = None
        self._scanner_manufacturer = slide.properties.get(PROPERTY_NAME_VENDOR)

    @property
    def magnification(self) -> Optional[float]:
        return self._magnification

    @property
    def scanner_manufacturer(self) -> Optional[str]:
        return self._scanner_manufacturer



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Dicomizer for bioformats source."""

from pathlib import Path
from typing import Optional, Sequence, Union

from PIL.Image import Image as PILImage
from pydicom import Dataset
from wsidicom import WsiDicom

from wsidicomizer.encoding import Encoder
from wsidicomizer.extras.bioformats.bioformats_source import BioformatsSource
from wsidicomizer.wsidicomizer import WsiDicomizer


class BioformatsDicomizer(WsiDicomizer):
    @classmethod
    def open(
        cls,
        filepath: Union[str, Path],
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        tile_size: int = 512,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        encoding_format: str = 'jpeg',
        encoding_quality: float = 90,
        jpeg_subsampling: str = '420',
        label: Optional[Union[PILImage, str, Path]] = None
    ) -> WsiDicom:
        """Open data in file in filepath as WsiDicom.

        Parameters
        ----------
        filepath: str
            Path to file
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
            Module datasets to use in files. If none, use default modules.
        tile_size: int = 512
            Tile size to use if not defined by file.
        include_levels: Optional[Sequence[int]] = None
            Optional list indices (in present levels) to include, e.g. [0, 1]
            includes the two lowest levels. Negative indicies can be used,
            e.g. [-1, -2] includes the two highest levels.
        include_label: bool = True
            Include label(s), default true.
        include_overwiew: bool = True
            Include overview(s), default true.
        include_confidential: bool = True
            Include confidential metadata.
        encoding_format: str = 'jpeg'
            Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'.
        encoding_quality: float = 90
            Quality to use if re-encoding. It is recommended to not use > 95 for jpeg.
            Use < 1 or > 1000 for lossless jpeg2000.
        jpeg_subsampling: str = '420'
            Subsampling option if using jpeg for re-encoding. Use '444' for
            no subsampling, '422' for 2x1 subsampling, and '420' for 2x2
            subsampling.
        label: Optional[Union[PILImage, str, Path]] = None
            Optional label image to use instead of label found in file.


        Returns
        ----------
        WsiDicom
            WsiDicom object of file.
        """
        if not isinstance(filepath, Path):
            filepath = Path(filepath)

        if not BioformatsSource.is_supported(filepath):
            raise NotImplementedError(f"{filepath} is not supported")
        encoder = Encoder.create_encoder(
            encoding_format,
            encoding_quality,
            subsampling=jpeg_subsampling
        )

        dicomizer = BioformatsSource(
            filepath,
            encoder,
            tile_size,
            modules,
            include_levels,
            include_label,
            include_overview,
            include_confidential
        )
        return cls(dicomizer, label)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Main module containing the WsiDicomizer class that allows non-DICOM files to be opened
like DICOM instances, enabling viewing and saving.
"""

import os
from pathlib import Path
from typing import Callable, List, Optional, Sequence, Type, Union

from PIL.Image import Image as PILImage
from pydicom.dataset import Dataset
from pydicom.uid import UID, generate_uid
from wsidicom import WsiDicom

from wsidicomizer.dicomizer_source import DicomizerSource
from wsidicomizer.encoding import Encoder
from wsidicomizer.sources import CziSource, OpenTileSource, TiffSlideSource

# List of supported Dicomizers in prioritization order.
loaded_sources: List[Type[DicomizerSource]] = [
    OpenTileSource,
    TiffSlideSource,
    CziSource,
]

try:
    from wsidicomizer.extras.openslide import OpenSlideSource

    loaded_sources.append(OpenSlideSource)
except ImportError:
    pass


class WsiDicomizer(WsiDicom):
    """Interface for Dicomizing files."""

    @classmethod
    def open(
        cls,
        filepath: Union[str, Path],
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        tile_size: int = 512,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        encoding_format: str = "jpeg",
        encoding_quality: float = 90,
        jpeg_subsampling: str = "420",
        label: Optional[Union[PILImage, str, Path]] = None,
        preferred_source: Optional[Type[DicomizerSource]] = None,
        **source_args,
    ) -> WsiDicom:
        """Open data in file in filepath as WsiDicom.

        Parameters
        ----------
        filepath: str
            Path to file
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
            Module datasets to use in files. If none, use default modules.
        tile_size: int = 512
            Tile size to use if not defined by file.
        include_levels: Optional[Sequence[int]] = None
            Optional list indices (in present levels) to include, e.g. [0, 1]
            includes the two lowest levels. Negative indicies can be used,
            e.g. [-1, -2] includes the two highest levels.
        include_label: bool = True
            Include label(s), default true.
        include_overwiew: bool = True
            Include overview(s), default true.
        include_confidential: bool = True
            Include confidential metadata.
        encoding_format: str = 'jpeg'
            Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'.
        encoding_quality: float = 90
            Quality to use if re-encoding. It is recommended to not use > 95 for jpeg.
            Use < 1 or > 1000 for lossless jpeg2000.
        jpeg_subsampling: str = '420'
            Subsampling option if using jpeg for re-encoding. Use '444' for
            no subsampling, '422' for 2x1 subsampling, and '420' for 2x2
            subsampling.
        label: Optional[Union[PILImage, str, Path]] = None
            Optional label image to use instead of label found in file.
        preferred_source: Optional[Type[DicomizerSource]] = None
            Optional override source to use.
        **source_args
            Optional keyword args to pass to source.

        Returns
        ----------
        WsiDicom
            WsiDicom object of file.
        """
        if not isinstance(filepath, Path):
            filepath = Path(filepath)

        selected_source = None
        if preferred_source is None:
            selected_source = next(
                (source for source in loaded_sources if source.is_supported(filepath)),
                None,
            )
        elif preferred_source.is_supported(filepath):
            selected_source = preferred_source
        if selected_source is None:
            raise NotImplementedError(f"{filepath} is not supported")
        encoder = Encoder.create_encoder(
            encoding_format, encoding_quality, subsampling=jpeg_subsampling
        )

        source = selected_source(
            filepath,
            encoder,
            tile_size,
            modules,
            include_levels,
            include_label,
            include_overview,
            include_confidential,
            **source_args,
        )
        return cls(source, label)

    @classmethod
    def convert(
        cls,
        filepath: Union[str, Path],
        output_path: Optional[Union[str, Path]] = None,
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None,
        tile_size: int = 512,
        uid_generator: Callable[..., UID] = generate_uid,
        include_levels: Optional[Sequence[int]] = None,
        include_label: bool = True,
        include_overview: bool = True,
        include_confidential: bool = True,
        workers: Optional[int] = None,
        chunk_size: Optional[int] = None,
        encoding_format: str = "jpeg",
        encoding_quality: float = 90,
        jpeg_subsampling: str = "420",
        offset_table: Optional[str] = "bot",
        label: Optional[Union[PILImage, str, Path]] = None,
        preferred_source: Optional[Type[DicomizerSource]] = None,
        **source_args,
    ) -> List[str]:
        """Convert data in file to DICOM files in output path. Created
        instances get UID from uid_generator. Closes when finished.

        Parameters
        ----------
        filepath: Union[str, Path],
            Path to file
        output_path: str = None
            Folder path to save files to.
        modules: Optional[Union[Dataset, Sequence[Dataset]]] = None
            Module datasets to use in files. If none, use default modules.
        tile_size: int = 512
            Tile size to use if not defined by file.
        uid_generator: Callable[..., UID] = generate_uid
             Function that can gernerate unique identifiers.
        include_levels: Optional[Sequence[int]] = None
            Optional list indices (in present levels) to include, e.g. [0, 1]
            includes the two lowest levels. Negative indicies can be used,
            e.g. [-1, -2] includes the two highest levels.
        include_label: bool = True
            Include label(s), default true.
        include_overwiew: bool = True
            Include overview(s), default true.
        include_confidential: bool = True
            Include confidential metadata.
        workers: Optional[int] = None,
            Maximum number of thread workers to use.
        chunk_size: Optional[int] = None,
            Chunk size (number of tiles) to process at a time. Actual chunk
            size also depends on minimun_chunk_size from image_data.
        encoding_format: str = 'jpeg'
            Encoding format to use if re-encoding. 'jpeg' or 'jpeg2000'.
        encoding_quality: float = 90
            Quality to use if re-encoding. It is recommended to not use > 95 for jpeg.
            Use < 1 or > 1000 for lossless jpeg2000.
        jpeg_subsampling: str = '420'
            Subsampling option if using jpeg for re-encoding. Use '444' for
            no subsampling, '422' for 2x1 subsampling, and '420' for 2x2
            subsampling.
        offset_table: Optional[str] = 'bot'
            Offset table to use, 'bot' basic offset table, 'eot' extended
            offset table, None - no offset table.
        label: Optional[Union[PILImage, str, Path]] = None
            Optional label image to use instead of label found in file.
        preferred_source: Optional[Type[DicomizerSource]] = None
            Optional override source to use.
        **source_args
            Optional keyword args to pass to source.

        Returns
        ----------
        List[str]
            List of paths of created files.
        """
        with cls.open(
            filepath,
            modules,
            tile_size,
            include_levels,
            include_label,
            include_overview,
            include_confidential,
            encoding_format,
            encoding_quality,
            jpeg_subsampling,
            label,
            preferred_source,
            **source_args,
        ) as wsi:
            if output_path is None:
                output_path = str(
                    Path(filepath).parents[0].joinpath(Path(filepath).stem)
                )
            try:
                os.mkdir(output_path)
            except FileExistsError:
                ValueError(f"Output path {output_path} already exists")
            created_files = wsi.save(
                output_path, uid_generator, workers, chunk_size, offset_table
            )

        return [str(filepath) for filepath in created_files]
